/*
** Copyright (C) 2003-2005 The Free Software Foundation.
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/


#include "libraries.h"
#include "utilities.h"
#include "export_utils.h"
#include "scripture.h"
#include "book.h"
#include "bible.h"
#include "usfm.h"
#include "usfmtools.h"
#include "project.h"
#include "sqlite3_plus.h"
#include "notes_utils.h"
#include "date_time_utils.h"


void export_get_markers_to_skip (set<string>& markers_to_skip)
{
  markers_to_skip.insert ("mt");
  markers_to_skip.insert ("mt1");
  markers_to_skip.insert ("mt2");
  markers_to_skip.insert ("s");
}

void export_to_bibleworks (Configuration * configuration, GtkWidget * parent, Progress * progress)
/*
Exports a whole project to a file that is fit for being imported by the 
BibleWorks Version Database Compiler.
This done in this manner:
- If an \id line is found, use its ID to get the name of the book in BibleWorks.
- If an chapter tag if found, use its number.
- As soon as an ID is found, put the current chapter at one.
- If a verse is found, add its number.
- Text of tags known not to be printable is thrown out.
- Any other tag's text gets added to the text of the verse.
Yes, this is a bit rough, I know...
*/
{
  // Get the filename where the data to save to.
  GtkWidget *dialog;
  dialog = gtk_file_chooser_dialog_new ("Save to a file", GTK_WINDOW (parent),
    GTK_FILE_CHOOSER_ACTION_SAVE, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
    GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT, NULL);
  gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (dialog), configuration->export_to_bibleworks_filename.c_str ());
  int result = gtk_dialog_run (GTK_DIALOG (dialog));
  string filename;
  if (result == GTK_RESPONSE_ACCEPT) 
    filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
  gtk_widget_destroy (dialog);  
  if (result == GTK_RESPONSE_ACCEPT) {
    // Store (new) filename.
    configuration->export_to_bibleworks_filename = filename;
    // Notebook shows progress bar.
    progress->start();
    // Start process.
    try
    {
      Project project (configuration->project);
      Usfm usfm (project.get_stylesheet());
      set<string> markers_to_skip;
      export_get_markers_to_skip (markers_to_skip);
      // Write to outputfile.
      WriteText wt (filename);
      // Open the Scripture and go through it.
      Scripture scripture (configuration->project);
      progress->set_iterate (0, 1, scripture.books.size());
      for (unsigned int bk = 0; bk < scripture.books.size(); bk++) {
        // Progress information.
        progress->iterate();
        if (progress->cancel) {
          progress->finish();
          return;
        }
        // Get the abbreviation to be used.
        string abbreviation;
        unsigned index = english_id_to_index (scripture.books[bk]);
        abbreviation = index_to_bibleworks_id (index);
        // Open the book
        Book book (scripture.paths[bk], false, "");
        // Start with chapter 1.
        string chapter = "1";
        // Go through the book.
        for (unsigned int i = 0; i < book.cleaned_lines.size(); i++) {
          ustring marker = usfm_extract_marker (book.cleaned_lines[i]);
          // Skip id line.
          if (usfm.is_identifier (marker))
            continue;
          // Skip running header.
          if (usfm.is_running_header (marker))
            continue;
          // Save chapter number.
          if (usfm.is_chapter_number(marker)) {
            chapter = number_in_string (book.cleaned_lines[i]);
            continue;
          }
          // If verse found, output new line, and its text.
          if (usfm.is_verse_number(marker)) {
            wt.out << endl << abbreviation << " " << chapter << ":" << book.cleaned_lines[i];
            continue;
          }
          // Skip list of know markers.
          if (markers_to_skip.find (marker) != markers_to_skip.end())
            continue;
          // Any other marker, known or unknown, output its text.
          wt.out << " " << book.cleaned_lines[i];
        }
      }
    }
    catch (exception & ex)
    {
      cerr << "Converting to BibleWorks: " << ex.what () << endl;
    }
  }
  // Notebook hides progress bar.
  progress->finish();
}


void export_translation_notes (const string& filename, ExportNotesFormat format, const vector<gint> & ids_to_display, bool export_all, GtkWidget * parent, Progress * progress)
// Exports the notes from the database.
/*
The exported xml file will look so:

<?xml version="1.0" encoding="UTF-8"?>
<bibledit-notes version="3">
  <note>
    <reference>
    </reference>
    <project>
    </project>
    <category>
    </category>
    <text>
    </text>
    <date-created>
    </date-created>
    <date-modified>
    </date-modified>
    <user>
    </user>
  </note>
</bibledit-notes>

This is subject to change as bibledit's notes system develop.
*/
{
  // Progress system.
  progress->start();
  progress->set_text ("Exporting notes ...");
  // Start process.
  try
  {
    // Write to outputfile.
    WriteText wt (filename);

    // Opening lines.
    switch (format) {
      case BibleditVersion3 :
      {
        wt.out << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" << endl;
        wt.out << "<bibledit-notes version=\"3\">" << endl;
        break;
      }
      case ScriptureNotesVersion20 :
      {
        wt.out << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" << endl;
        wt.out << "<ScriptureNotes Version=\"2.0\">" << endl;
        break;
      }
    }

    // Connect to database.
    sqlite3::connection con (notes_database_filename ().c_str ());

    // Get the number of notes.
    gint notes_count;
    notes_count = con.executeint32("select count(*) from '%s';", TABLE_NOTES);
    progress->set_iterate (0, 1, notes_count);
    gint note_counter = 0;

    // Exporting all or selection?
    set<gint> ids (ids_to_display.begin(), ids_to_display.end());
   
    // Go through all the notes.
    sqlite3::reader reader;
    reader = con.executereader ("select reference, project, category, note, created, modified, user, id from '%s';", TABLE_NOTES);
    while (reader.read()) {
      // Progress information.
      note_counter++;
      progress->iterate();
      if (progress->cancel) {
        progress->finish();
        return;
      }
      
      // See whether to display this particular id.
      if (!export_all) {
        gint id = reader.getint32 (7);
        if (ids.find (id) == ids.end())
          continue;
      }
      
      // Start note.
      switch (format) {
        case BibleditVersion3 :
        {
          wt.out << "  <note>" << endl;
          break;
        }
        case ScriptureNotesVersion20 :
        {
          wt.out << "  <Note>" << endl;
          break;
        }
      }
      
      // Deal with the references.
      string reference = reader.getstring(0);
      // Parse the string into its possible several references.
      Parse parse (reference);
      reference.clear();
      switch (format) {
        case BibleditVersion3 :
        {
          // Go through each reference.
          for (unsigned int i = 0; i < parse.words.size(); i++) {
            // Make it human readable.
            int book, chapter, verse;
            numerical_equivalent_to_reference (parse.words[i], book, chapter, verse);
            if (!reference.empty())
              reference.append ("\n");
            reference.append (book_chapter_verse_to_reference (book, chapter, verse));
          }
          wt.out << "    <references>" << reference << "</references>" << endl;
          break;
        }
        case ScriptureNotesVersion20 :
        {
          // Go through each reference.
          for (unsigned int i = 0; i < parse.words.size(); i++) {
            // Make it human readable.
            int book, chapter, verse;
            numerical_equivalent_to_reference (parse.words[i], book, chapter, verse);
            if (!reference.empty())
              reference.append ("\n");
            reference.append (book_chapter_verse_to_reference_paratext (book, chapter, verse));
          }
          wt.out << "    <References>" << reference << "</References>" << endl;
          break;
        }
      }
      
      // Deal with the project name.
      ustring project = reader.getstring(1);
      switch (format) {
        case BibleditVersion3 :
        {
          wt.out << "    <project>" << project << "</project>" << endl;
          break;
        }
        case ScriptureNotesVersion20 :
        {
          wt.out << "    <Project>" << project << "</Project>" << endl;
          break;
        }
      }
      
      // Deal with the status of the note.
      string category = reader.getstring(2);
      switch (format) {
        case BibleditVersion3 :
        {
          string referred_to;
          wt.out << "    <category>" << category << "</category>" << endl;
          break;
        }
        case ScriptureNotesVersion20 :
        {
          string tne_status;
          string category_lowercase = lowerCase (category);
          if (category_lowercase == "no issue")
            tne_status = "No Issue";
          else if (category_lowercase == "unresolved")
            tne_status = "Unresolved";
          else if (category_lowercase == "resolved")
            tne_status = "Resolved";
          else
            tne_status = "Unresolved";
          wt.out << "    <Status>" << tne_status << "</Status>" << endl;
          break;
        }
      }
      
      // The note text.
      ustring note = reader.getstring(3);
      switch (format) {
        case BibleditVersion3 :
        {
          wt.out << "    <text>" << note << "</text>" << endl;
          break;
        }
        case ScriptureNotesVersion20 :
        {
          wt.out << "    <Contents><P>" << note << "</P></Contents>" << endl;
          break;
        }
      }

      // The date it was created.
      int created = reader.getint32(4);
      string date;
      date = trim (date_time_julian_human_readable (created, false));
      switch (format) {
        case BibleditVersion3 :
        {
          wt.out << "    <date-created>" << date << "</date-created>" << endl;
          break;
        }
        case ScriptureNotesVersion20 :
        {
          wt.out << "    <CreationDate>" << date << "</CreationDate>" << endl;
          break;
        }
      }

      // Deal with the modification date.
      int modified = reader.getint32(5);
      date = trim (date_time_julian_human_readable (modified, false));
      switch (format) {
        case BibleditVersion3 :
        {
          wt.out << "    <date-modified>" << date << "</date-modified>" << endl;
          break;
        }
        case ScriptureNotesVersion20 :
        {
          wt.out << "    <ModificationDate>" << date << "</ModificationDate>" << endl;
          break;
        }
      }

      // Mention the one who created the note.
      ustring user = reader.getstring(6);
      switch (format) {
        case BibleditVersion3 :
        {
          wt.out << "    <created-by>" << user << "</created-by>" << endl;
          break;
        }
        case ScriptureNotesVersion20 :
        {
          wt.out << "    <CreatedBy>" << user << "</CreatedBy>" << endl;
          break;
        }
      }

      // Finish note.
      switch (format) {
        case BibleditVersion3 :
        {
          wt.out << "  </note>" << endl;
          break;
        }
        case ScriptureNotesVersion20 :
        {
          wt.out << "  </Note>" << endl;
          break;
        }
      }
    }
    // Closing lines.
    switch (format) {
      case BibleditVersion3 :
      {
        wt.out << "</bibledit-notes>" << endl;
        break;
      }
      case ScriptureNotesVersion20 :
      {
        wt.out << "</ScriptureNotes>" << endl;
        break;
      }
    }
    // Close connection.  
    reader.close();      
    con.close ();
  }
  catch (exception & ex)
  {
    cerr << "Exporting notes: " << ex.what () << endl;
  }
    
  // End progress system.
  progress->finish();
}
