/*
  ** Copyright (C) 2003-2005 The Free Software Foundation.
  **  
  ** This program is free software; you can redistribute it and/or modify
  ** it under the terms of the GNU General Public License as published by
  ** the Free Software Foundation; either version 2 of the License, or
  ** (at your option) any later version.
  **  
  ** This program is distributed in the hope that it will be useful,
  ** but WITHOUT ANY WARRANTY; without even the implied warranty of
  ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  ** GNU General Public License for more details.
  **  
  ** You should have received a copy of the GNU General Public License
  ** along with this program; if not, write to the Free Software
  ** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  **  
  */


#include "libraries.h"
#include "utilities.h"
#include <libgen.h>
#include <glib.h>
#include <config.h>
#include "notes_utils.h"
#include "directories.h"
#include "sqlite3_plus.h"
#include "bible.h"
#include "date_time_utils.h"
#include "log.h"
#include "gwrappers.h"


#define TABLE_VERSION "version"
#define NOTES_DATABASE_VERSION 3


void notes_database_verify_update ()
// Updates the notes database to the current version.
{
  try
  {
    // Get the version number of the database.
    int version;
    {
      // Connect to the database.
      sqlite3::connection con (notes_database_filename ().c_str ());
      // How many tables are there?
      int count;
      count = con.executeint32("select count(*) from sqlite_master;");
      if (count == 1) {
        // If there is 1 table, then the version number is 1, because that
        // was the structure of the notes database at the time of version 1.
        version = 1;
      } else {
        // If there are more than one table, get the version number from the
        // database.
        sqlite3::reader reader;
        reader = con.executereader ("select version from '%s';", TABLE_VERSION);
        if (reader.read ()) {
          version = reader.getint32(0);
        }
        reader.close ();
      }
      con.close ();
      // Log database version.
      log_message ("Notes database version " + convert_to_string (version));
    }
    
    // Update the database to the newest version. Using the fall-through
    // mechanism of the switch statement, the database is updated automatically
    // to the newest version, step by step.
    // Note that the ALTER TABLE command is not fully implemented in sqlite3,
    // and the command to add a row was implemented somewhere in version 3,
    // so we cannot rely on it.
    string original_database_name = notes_database_filename ();
    string temporal_database_name = original_database_name + "_temporal";
    switch (version) {
      case 0 : 
      {
      }
      case 1 :
      {
        // Upgrade to version 2.
        log_message ("Upgrading notes database to version 2");
        
        GtkWidget *updatedatabasewindow;
        GtkWidget *vbox1;
        GtkWidget *label1;
        GtkWidget *progressbar1;
        updatedatabasewindow = gtk_window_new (GTK_WINDOW_TOPLEVEL);
        gtk_window_set_position (GTK_WINDOW (updatedatabasewindow), GTK_WIN_POS_CENTER);
        gtk_window_set_default_size (GTK_WINDOW (updatedatabasewindow), 400, -1);
        gtk_window_set_type_hint (GTK_WINDOW (updatedatabasewindow), GDK_WINDOW_TYPE_HINT_SPLASHSCREEN);
        vbox1 = gtk_vbox_new (FALSE, 0);
        gtk_widget_show (vbox1);
        gtk_container_add (GTK_CONTAINER (updatedatabasewindow), vbox1);
        label1 = gtk_label_new ("Bibledit is updating the notes to version 2");
        gtk_widget_show (label1);
        gtk_box_pack_start (GTK_BOX (vbox1), label1, FALSE, FALSE, 0);
        progressbar1 = gtk_progress_bar_new ();
        gtk_widget_show (progressbar1);
        gtk_box_pack_start (GTK_BOX (vbox1), progressbar1, FALSE, FALSE, 0);
        gtk_widget_show (updatedatabasewindow);
        int updated_records = 0;        
        // Read values from old notes database, store in memory.
        vector<int> id;
        vector<string> reference;
        vector<string> project;
        vector<int> referred;
        vector<string> note;
        vector<string> casefolded;
        vector<int> created;
        vector<int> modified;
        vector<string> user;
        {
          sqlite3::connection con (original_database_name.c_str());
          sqlite3::reader reader;
          reader = con.executereader ("select id, reference, project, referred, note, casefolded, created, modified, user from %s;", TABLE_NOTES);
          while (reader.read ()) {
            updated_records++;
            id.push_back (reader.getint32(0));
            reference.push_back (reader.getstring(1));
            project.push_back (double_apostrophy (reader.getstring(2)));
            referred.push_back (reader.getint32(3));
            note.push_back (double_apostrophy (reader.getstring(4)));
            casefolded.push_back (double_apostrophy (reader.getstring(5)));
            created.push_back (reader.getint32(6));
            modified.push_back (reader.getint32(7));
            user.push_back (reader.getstring(8));
          }
          reader.close ();
          con.close ();
        }
        {
          // Create temporal new database.
          sqlite3::connection con (temporal_database_name.c_str ());
          con.executenonquery("create table '%s' (id integer, reference text, project text, referred integer, note text, casefolded text, created integer, modified integer, user text);", TABLE_NOTES);
          con.executenonquery("create table '%s' (version integer);", TABLE_VERSION);
          con.executenonquery("insert into '%s' values (%d);", TABLE_VERSION, 2);
          // Store all values in it.
          for (unsigned int i = 0; i < id.size(); i++) {
            double fraction;
            fraction = double (i) / double (updated_records);
            gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (progressbar1), fraction);
            while (gtk_events_pending ()) gtk_main_iteration ();
            con.executenonquery ("insert into %s values (%d, '%s', '%s', %d, '%s', '%s', %d, %d, '%s');", TABLE_NOTES, id[i], reference[i].c_str(), project[i].c_str(), referred[i], note[i].c_str(), casefolded[i].c_str(), created[i], modified[i], user[i].c_str());
          }
          con.close ();
        }

        // Copy temporal database back to the database.
        unlink (original_database_name.c_str());
        string command;
        command = "cp '" + temporal_database_name + "' '" + original_database_name + "'";
        system (command.c_str());
        unlink (temporal_database_name.c_str());
        
        gtk_widget_destroy (updatedatabasewindow);

        // Log.
        log_message ("Records updated: " + convert_to_string(updated_records));
      }
      case 2 :
      {
        // Upgrade to version 3.
        log_message ("Upgrading notes database to version 3");
        
        GtkWidget *updatedatabasewindow;
        GtkWidget *vbox1;
        GtkWidget *label1;
        GtkWidget *progressbar1;
        updatedatabasewindow = gtk_window_new (GTK_WINDOW_TOPLEVEL);
        gtk_window_set_position (GTK_WINDOW (updatedatabasewindow), GTK_WIN_POS_CENTER);
        gtk_window_set_default_size (GTK_WINDOW (updatedatabasewindow), 400, -1);
        gtk_window_set_type_hint (GTK_WINDOW (updatedatabasewindow), GDK_WINDOW_TYPE_HINT_SPLASHSCREEN);
        vbox1 = gtk_vbox_new (FALSE, 0);
        gtk_widget_show (vbox1);
        gtk_container_add (GTK_CONTAINER (updatedatabasewindow), vbox1);
        label1 = gtk_label_new ("Bibledit is updating the notes to version 3");
        gtk_widget_show (label1);
        gtk_box_pack_start (GTK_BOX (vbox1), label1, FALSE, FALSE, 0);
        progressbar1 = gtk_progress_bar_new ();
        gtk_widget_show (progressbar1);
        gtk_box_pack_start (GTK_BOX (vbox1), progressbar1, FALSE, FALSE, 0);
        gtk_widget_show (updatedatabasewindow);
        int updated_records = 0;        
        // Read values from old notes database, store in memory.
        vector<int> id;
        vector<string> reference;
        vector<string> project;
        vector<int> referred;
        vector<string> note;
        vector<string> casefolded;
        vector<int> created;
        vector<int> modified;
        vector<string> user;
        {
          sqlite3::connection con (original_database_name.c_str());
          sqlite3::reader reader;
          reader = con.executereader ("select id, reference, project, referred, note, casefolded, created, modified, user from %s;", TABLE_NOTES);
          while (reader.read ()) {
            updated_records++;
            id.push_back (reader.getint32(0));
            reference.push_back (reader.getstring(1));
            project.push_back (double_apostrophy (reader.getstring(2)));
            referred.push_back (reader.getint32(3));
            note.push_back (double_apostrophy (reader.getstring(4)));
            casefolded.push_back (double_apostrophy (reader.getstring(5)));
            created.push_back (reader.getint32(6));
            modified.push_back (reader.getint32(7));
            user.push_back (reader.getstring(8));
          }
          reader.close ();
          con.close ();
        }
        {
          // Create temporal new database.
          sqlite3::connection con (temporal_database_name.c_str ());
          con.executenonquery("create table '%s' (id integer, reference text, project text, category text, note text, casefolded text, created integer, modified integer, user text);", TABLE_NOTES);
          con.executenonquery("create table '%s' (version integer);", TABLE_VERSION);
          con.executenonquery("insert into '%s' values (%d);", TABLE_VERSION, 3);
          // Store all values in it.
          for (unsigned int i = 0; i < id.size(); i++) {
            double fraction;
            fraction = double (i) / double (updated_records);
            gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (progressbar1), fraction);
            while (gtk_events_pending ()) gtk_main_iteration ();
            ustring category;
            switch (referred[i]) {
              case 1:  { category = "For myself"; break; }
              case 2:  { category = "For subteam"; break; }
              case 3:  { category = "For team"; break; }
              case 4:  { category = "For panel"; break; }
              case 5:  { category = "For church"; break; }
              case 6:  { category = "For consultant"; break; }
              case 7:  { category = "For Bible society"; break; }
              default: { category = "No issue"; break; }
            }
            con.executenonquery ("insert into %s values (%d, '%s', '%s', '%s', '%s', '%s', %d, %d, '%s');", TABLE_NOTES, id[i], reference[i].c_str(), project[i].c_str(), category.c_str(), note[i].c_str(), casefolded[i].c_str(), created[i], modified[i], user[i].c_str());
          }
          con.close ();
        }

        // Copy temporal database back to the database.
        unlink (original_database_name.c_str());
        string command;
        command = "cp '" + temporal_database_name + "' '" + original_database_name + "'";
        system (command.c_str());
        unlink (temporal_database_name.c_str());
        
        gtk_widget_destroy (updatedatabasewindow);

        // Log.
        log_message ("Records updated: " + convert_to_string(updated_records));
      }
      case 3 :
      {
        // Up-to-date.
        log_message ("Notes database is up to date");
      }
    }
  }
  catch (exception & ex)
  {
    cerr << "Updating notes database: " << ex.what () << endl;
  }
}


void notes_database_verify_create ()
/*
This verifies the notes database is fine.
If it's not there, it creates one.
*/
{
  try
  {
    // Connect to the database.
    sqlite3::connection con (notes_database_filename ().c_str ());
    int count;
    // Check for the table.
    count = con.executeint32("select count(*) from sqlite_master where name='%s';", TABLE_NOTES);
    if (count == 0) {
      // Create the notes table.
      con.executenonquery("create table '%s' (id integer, reference text, project text, category text, note text, casefolded text, created integer, modified integer, user text);", TABLE_NOTES);
      // Create the version table and store version number.
      con.executenonquery("create table '%s' (version integer);", TABLE_VERSION);
      con.executenonquery("insert into '%s' values (%d);", TABLE_VERSION, NOTES_DATABASE_VERSION);
    }
    con.close ();
    notes_database_verify_update ();
  }
  catch (exception & ex)
  {
    cerr << "Initializing notes database: " << ex.what () << endl;
  }
}


string notes_database_filename ()
// Returns the filename of the notes database.
{
  return gw_build_filename (directories_get_notes (), "database");
}


void insert_link (GtkTextBuffer *buffer, string text, gint id)
/* Inserts a piece of text into the buffer, giving it the usual
 * appearance of a hyperlink in a web browser: blue and underlined.
 * Additionally, attaches some data on the tag, to make it recognizable
 * as a link. 
 */
{
  GtkTextTag *tag;
  tag = gtk_text_buffer_create_tag (buffer, NULL, 
				    "foreground", "blue", 
				    "underline", PANGO_UNDERLINE_SINGLE, 
				    NULL);
  g_object_set_data (G_OBJECT (tag), "id", GINT_TO_POINTER (id));
  GtkTextIter iter;
  gtk_text_buffer_get_end_iter (buffer, &iter);
  gtk_text_buffer_insert_with_tags (buffer, &iter, text.c_str(), -1, tag, NULL);
}
  

gint notes_database_get_unique_id ()
// This generates a unique id, one that is not yet used in the notes database.
{
  gint32 result;
  try
  {
    sqlite3::connection con (notes_database_filename ().c_str ());
    bool found = false;
    while (!found) {
      result = g_random_int_range (1, 100000000);
      gint count = con.executeint32("select count(*) from '%s' where id=%i;", TABLE_NOTES, result);
      if (count == 0) {
        found = true;
      }
    }
    con.close ();
  }
  catch (exception & ex)
  {
    cerr << "Getting random id for notes database: " << ex.what () << endl;
  }
  return result;
}


void notes_delete_one (int id)
// Deletes the note with id.
{
  try
  {
    // Connect to database and start transaction
    sqlite3::connection con (notes_database_filename ().c_str ());
    con.executenonquery ("begin;");
    // Delete data with "id".
    con.executenonquery ("delete from %s where id = %d;", TABLE_NOTES, id);
    // Commit the transaction and close connection.
    con.executenonquery ("commit;");
    con.close ();
  }
  catch (exception & ex)
  {
    cerr << "Deleting note: " << ex.what () << endl;
  }
}


void notes_sort (vector<gint32>& ids, const vector<string>& refs, const vector<int>& dates)
/*
This sorts notes.
ids         - ID's of the notes.
              The ids will be modified and sorted, so that these can be used 
              to fetch data from the database in the right order.
refs        - References to sort on first.
dates       - Modification dates to sort on next.
*/
{
  // No sorting when less than two entries.
  if (ids.size() < 2)
    return;

  // The sorting is done by putting everything in one string, and then
  // sorting the vector of strings, and then taking out the ids again.

  // Storage for strings to sort
  vector<string> strings_to_sort;
  
  // Go through all data to sort.
  for (unsigned int i = 0; i < ids.size(); i++) {
  
    // Storage for string to build.
    string string_to_build;
    
    // First part of the string is the reference.
    // It always has the same length, so can be taken as it.
    string_to_build = refs[i];

    // Next part to sort on is the date. This Julian day has a length of 6
    // characters at the time of programming, and will continue so until
    // Deo Volente the year 2739 A.D. Therefore there is no need to modify the
    // length of it.
    {
      string date;
      date = convert_to_string (dates[i]);
      string_to_build.append(date);
    }
    
    // Last part of the string to sort is the id.
    // Maximum id = 100000000, so maximum length of 9 characters.
    {
      string id;
      id = "000000000" + convert_to_string (ids[i]);
      id = id.substr(id.length() - 9, 9);
      string_to_build.append (id);
    }

    // Store built string
    strings_to_sort.push_back(string_to_build);    
  }

  // Sort the numerical equivalents.
  sort (strings_to_sort.begin(), strings_to_sort.end());
  
  // Extract the sorted ids from the sorted data.
  ids.clear();
  for (unsigned int i = 0; i < strings_to_sort.size(); i++) {
    string id = strings_to_sort[i].substr(15, 9);
    gint32 numerical_id = convert_to_int (id);
    ids.push_back(numerical_id);
  }
}


void notes_select (Configuration * configuration, vector<gint32>& ids, const string& currentreference)
/*
This selects notes for display.
It does this based on the criteria passed.
The resulting selection will be stored in ids.
*/
{
  // Clear ids.
  ids.clear();
  try
  {
    // Connect to database.
    sqlite3::connection con (notes_database_filename ().c_str ());
    // First read from database, for sorting purposes.
    sqlite3::reader reader;
    // See which notes to select.
    switch (configuration->show_notes_selector) {
      case snsCurrentVerse:
      {
        string book, chapter, verse;
        decode_reference (currentreference, book, chapter, verse);
        string this_verse;
        this_verse = reference_to_numerical_equivalent (book, chapter, verse);      
        reader = con.executereader ("select id, reference, modified, project from '%s' where reference glob ('*%s*');", TABLE_NOTES, this_verse.c_str());
        break;
      }
      case snsToday:
      {
        int currentdate = date_time_julian_day_get_current ();
        reader = con.executereader ("select id, reference, modified, project from '%s' where modified = %d;", TABLE_NOTES, currentdate);
        break;
      }
      case snsAll:
      {
        reader = con.executereader ("select id, reference, modified, project from '%s';", TABLE_NOTES);
        break;
      }
      case snsCategory:
      {
        reader = con.executereader ("select id, reference, modified, project from '%s' where category = '%s';", TABLE_NOTES, configuration->show_notes_category.c_str());
        break;
      }
      case snsDateRange:
      {
        reader = con.executereader ("select id, reference, modified, project from '%s' where modified  >= %d and modified <= %d;", TABLE_NOTES, configuration->show_notes_date_range_from, configuration->show_notes_date_range_to);
        break;
      }
    }
    // Storage for sorting purposes.
    vector<string> references;
    vector<int> dates;
    // Read all resulting data from the db.
    while (reader.read ()) {
      // See whether we show notes for current project only.
      if (configuration->show_notes_for_current_project_only) {
        string project_in_db;
        project_in_db = reader.getstring(3);
        if (project_in_db != configuration->project)
          // Project is different: do not show the notes.
          continue;
      }
      // Get id.
      gint32 id = reader.getint32(0);
      // Get the numerical equivalent of the reference.
      string reference;
      reference = reader.getstring (1);
      {
        // Parse the string into its possible several references.
        Parse parse (reference);
        // Take the first refence available.
        if (parse.words.size() > 0)
          reference = parse.words[0];
      }
      // Get date modified.
      int date;
      date = reader.getint32(2);
      // Store data.
      ids.push_back(id);
      references.push_back(reference);
      dates.push_back(date);
    }
    // Close database
    reader.close ();
    con.close ();
    // Sort the notes.
    notes_sort (ids, references, dates);
  }
  catch (exception & ex)
  {
    cerr << "Error: " << ex.what () << endl;
  }
}


void notes_display_one (GtkWidget *textview, GtkTextBuffer *textbuffer, gint32 id)
/*
This cares for the actual displaying of the notes.
It displays the note with an ID of id.
*/
{
  try
  {
    // Connect to database.
    sqlite3::connection con (notes_database_filename ().c_str ());
    // We've got an id display here. 
    
    // Get a reader for the database.
    sqlite3::reader reader;
    // Execute the SQL statement.
    reader = con.executereader ("select id, reference, note from '%s' where id = %d;", TABLE_NOTES, id);
    // Go through the results (there should be only one anyway).
    while (reader.read ()) {

      // Get the numerical equivalent of the reference.
      string reference;
      reference = reader.getstring (1);

      // Parse the string into its possible several references.
      Parse parse (reference);
      reference.clear();

      // Go through each reference.
      for (unsigned int i2 = 0; i2 < parse.words.size(); i2++) {
        // Make it human readable.
        int book, chapter, verse;
        numerical_equivalent_to_reference (parse.words[i2], book, chapter, verse);
        if (!reference.empty())
          reference.append (", ");
        reference.append (book_chapter_verse_to_reference (book, chapter, verse));
      }

      // Store position of cursor.
      gint linenumber, offset;
      {
        GtkTextIter cursorposition;
        GtkTextMark * textmark;
        textmark = gtk_text_buffer_get_insert (textbuffer);
        gtk_text_buffer_get_iter_at_mark (textbuffer, &cursorposition, textmark);
        linenumber = gtk_text_iter_get_line (&cursorposition);
        offset = gtk_text_iter_get_line_offset (&cursorposition);
      }

      // Move cursor to end of buffer.
      GtkTextIter endposition;
      gtk_text_buffer_get_end_iter (textbuffer, &endposition);
      gtk_text_buffer_place_cursor (textbuffer, &endposition);
      
      // Insert a link with this heading.
      insert_link (textbuffer, reference, id); 
      gtk_text_buffer_insert_at_cursor (textbuffer, "\n", -1);
        
      // Get the text of the note.
      ustring note;
      note = reader.getstring (2);
      gtk_text_buffer_insert_at_cursor (textbuffer, note.c_str(), -1);
      gtk_text_buffer_insert_at_cursor (textbuffer, "\n", -1);
      
      // Move cursor to where it was before.
      GtkTextIter formerposition;
      gtk_text_buffer_get_iter_at_line_offset (textbuffer, &formerposition, linenumber, offset);
      gtk_text_buffer_place_cursor (textbuffer, &formerposition);
        
    }
    // Close reader.
    reader.close ();
      
    // Close connection
    con.close ();
  }
  catch (exception & ex)
  {
    cerr << "Error: " << ex.what () << endl;
  }
}


void notes_get_references_from_editor (GtkTextBuffer *textbuffer, vector<string>& references, vector<string>& messages)
/*
Gets all references from the notes editor.
Normalizes them.
Produces messages on trouble.
*/
{
  // Get all lines from the textbuffer.
  vector<ustring> lines;
  textbuffer_get_lines (textbuffer, lines);
  // When discovering a reference from a user's entry, use previous values,
  // so that it becomes quicker for a user to enter new references.
  // If Leviticus 10:11 is already there, and the user wishes to add verse
  // 12 also, he just enters 12 on a line, and that' it.
  string previousbook, previouschapter, previousverse;
  for (unsigned int i = 0; i < lines.size(); i++) {
    if (!lines[i].empty ()) {
      // Normalize reference.
      string book;
      string chapter;
      string verse;
      if (reference_discover (previousbook, previouschapter, previousverse, lines[i], book, chapter, verse)) {
        // Store reference.
        string ref = book + " " + chapter + ":" + verse;
        references.push_back (ref);
        // Store values to discover next reference.
        previousbook = book;
        previouschapter = chapter;
        previousverse = verse;
      } else {
        messages.push_back ("Reference " + lines[i] + " is not valid and has been removed.");
      }
    }
  }
}


string notes_categories_filename ()
// Returns the filename of the notes database.
{
  return gw_build_filename (directories_get_notes (), "categories");
}


void notes_categories_check ()
// Check categories are there - if not, create default set.
{
  if (!g_file_test (notes_categories_filename ().c_str(), G_FILE_TEST_IS_REGULAR)) {
    vector<string> categories;
    categories.push_back ("No issue");
    categories.push_back ("For myself");
    categories.push_back ("For subteam");
    categories.push_back ("For team");
    categories.push_back ("For panel");
    categories.push_back ("For church");
    categories.push_back ("For consultant");
    categories.push_back ("For Bible society");
    write_lines (notes_categories_filename (), categories);
  }
}


void notes_categories_add_from_database (vector<string> & categories)
// Takes the existing notes categories, if there are any, and adds any
// extra categories found in the database.
{
  try
  {
    // Get the unique categories.
    set<string> database_categories;
    sqlite3::connection con (notes_database_filename ().c_str ());
    sqlite3::reader reader;
    reader = con.executereader ("select category from '%s';", TABLE_NOTES);
    while (reader.read ()) {
      database_categories.insert (reader.getstring (0));
    }
    reader.close ();
    con.close ();
    // Add any new categories to the container.
    set<string> localcategories (categories.begin(), categories.end());
    vector<string> db_categories (database_categories.begin(), database_categories.end());
    for (unsigned int i = 0; i < db_categories.size(); i++) {
      if (localcategories.find (db_categories[i]) == localcategories.end()) {
        categories.push_back (db_categories[i]);
      }
    }
  }
  catch (exception & ex)
  {
    cerr << "Error getting notes categories: " << ex.what () << endl;
  }
}


void notes_database_vacuum ()
{
  try
  {
    sqlite3::connection con (notes_database_filename().c_str ());
    con.executenonquery("vacuum;");
    con.close ();
  }
  catch (exception & ex)
  {
    cerr << "Error vacuuming notes: " << ex.what () << endl;
  }
}
