/*
** Copyright (C) 2003-2005 The Free Software Foundation.
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/


#include "utilities.h"
#include "search_utils.h"
#include "worddatabase.h"
#include "sqlite3_plus.h"
#include "bible.h"


void search_string_internal (const ustring& searchfor, bool casesensitive, 
                             const ustring& project, vector<string>& results,
                             bool use_book_selection, set<string>& books)
{
  try
  {
    // Connect to database.
    Worddatabase worddatabase (0);
    sqlite3::connection con (worddatabase.dbfilename (project).c_str ());
    // Get the string to search for. Note any apostrophies need to be doubled for SQLite.
    // We need to normalize the search expression, as prescribed, when comparing strings.
    ustring localsearchword;
    localsearchword = searchfor;
    localsearchword = localsearchword.normalize ();
    if (!casesensitive)
      localsearchword = localsearchword.casefold ();
    localsearchword = double_apostrophy (localsearchword);
    /*
    There is an additional complication because we use the SQL glob function
    to get the results quickly from the database. But this implies that the
    characters that are used in Unix globbing work different from what is
    expected.
    These characters are four: * ? [ ]
    Measures are taken so that these four characters behave as should, that is,
    searching for * really searches for the character * and does not work like
    in Unix filename globbing. The same thing for the other characters.
    The problem is solved by looking whether the user searches for one of these
    characters. If so, all data is retrieved from the database, and a manual
    search is performed on the results. It slows down the process a little.
    */
    // Open a reader.
    sqlite3::reader reader;
    if (localsearchword.find_first_of ("*?[]") != string::npos) {
      // Ok, any of the special characters are present. Slower search.
      // Execute the SQL statement, and read the results.
      if (casesensitive) {
        reader = con.executereader ("select book, chapter, verse, line from 'lines';");
      } else {
        reader = con.executereader ("select book, chapter, verse, casefolded from 'lines';");
      }
      while (reader.read ()) {
        ustring line = reader.getstring (3);
        if (line.find (localsearchword) != string::npos) {
          if (use_book_selection) {
            string book = index_to_biblebook (reader.getint32 (0));
            if (books.find (book) != books.end ()) {
              results.push_back (book_chapter_verse_to_reference (reader.getint32 (0), reader.getint32 (1), reader.getint32 (2)));
            }
          } else {
            results.push_back (book_chapter_verse_to_reference (reader.getint32 (0), reader.getint32 (1), reader.getint32 (2)));
          }
        }
      }
    } else {
      // Ok, no special characters: fast search.
      // Execute the SQL statement, and read the results.
      if (casesensitive) {
        reader = con.executereader ("select book, chapter, verse from 'lines' where line glob ('*%s*');", localsearchword.c_str ());
      } else {
        reader = con.executereader ("select book, chapter, verse from 'lines' where casefolded glob ('*%s*');", localsearchword.c_str ());
      }
      while (reader.read ()) {
        if (use_book_selection) {
          string book = index_to_biblebook (reader.getint32 (0));
          if (books.find (book) != books.end ()) {
            results.push_back (book_chapter_verse_to_reference (reader.getint32 (0), reader.getint32 (1), reader.getint32 (2)));
          }
        } else {
          results.push_back (book_chapter_verse_to_reference (reader.getint32 (0), reader.getint32 (1), reader.getint32 (2)));
        }
      }
    }
    // Close all
    reader.close ();
    con.close ();
  }
  catch (exception & ex)
  {
    cerr << "Searching: " << ex.what () << endl;
  }
}


void search_string (const ustring& searchfor, bool casesensitive, 
                    const ustring& project, vector<string>& results)
{
  set<string> dummy;
  search_string_internal (searchfor, casesensitive, project, results, false, dummy);
}


void search_string (const ustring& searchfor, bool casesensitive, 
                    const ustring& project, set <string>& books,
                    vector<string>& results)
{
  search_string_internal (searchfor, casesensitive, project, results, true, books);
}
