/*
** Copyright (C) 2003-2005 The Free Software Foundation.
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/


#include "libraries.h"
#include "utilities.h"
#include "stylesheetutils.h"
#include "directories.h"
#include <config.h>
#include "constants.h"
#include "xmlutils.h"
#include "gwrappers.h"


void stylesheet_get_ones_available (vector<string>& names)
// Gets the names of the stylesheets that are there.
{
  ReadDirectories rd (directories_get_stylesheets (), "", "");
  names.assign (rd.directories.begin(), rd.directories.end());
}


void stylesheet_create_new (const string& name)
// Create a new stylesheet from templates, named "name".
{
  string directory;
  directory = directories_get_stylesheets ();
  directory = gw_build_filename (directory, name);
  create_directory (directory);
  ReadFiles rf (PACKAGE_DATA_DIR, "", ".style");
  for (unsigned int i = 0; i < rf.files.size(); i++) {
    string command;
    command = "cp ";
    string file = gw_build_filename (PACKAGE_DATA_DIR, rf.files[i]);
    command.append (file);
    command.append (" '");
    file = rf.files[i];
    file = file.substr (0, file.length() - 6);
    file.append (".xml");
    file = gw_build_filename (directory, file);
    command.append (file);
    command.append ("'");
    system(command.c_str());
  }
}


void stylesheet_delete (const string& name)
// Deletes a stylesheet.
{
  string directory;
  directory = directories_get_stylesheets ();
  directory = gw_build_filename (directory, name);
  string command;
  command = "rm -rf '";
  command.append (directory);
  command.append ("'");
  system(command.c_str());
}


void stylesheet_copy (const string& from_name, const string& to_name)
// Copies one stylesheet to another.
{
  string from_dir;
  from_dir = directories_get_stylesheets ();
  from_dir = gw_build_filename (from_dir, from_name);
  string to_dir;
  to_dir = directories_get_stylesheets ();
  to_dir = gw_build_filename (to_dir, to_name);
  string command;
  command = "cp -r '";
  command.append (from_dir);
  command.append ("' '");
  command.append (to_dir);
  command.append ("'");
  system (command.c_str());
}


void stylesheet_get_styles (const string& stylesheet, vector<Style>& styles)
// Get all data of all stylesheets.
{
  // When a project uses a stylesheet that does not exist, use another one,
  // or create one.
  // No longer done so, but the Project object checks on that.
  string directory;
  directory = directories_get_stylesheets ();
  directory = gw_build_filename (directory, stylesheet);
  ReadFiles rf (directory, "", ".xml");
  for (unsigned int i = 0; i < rf.files.size(); i++) {
    Style style (directory, rf.files[i].substr (0, rf.files[i].length() - 4), false);
    styles.push_back(style);
  }
}


void stylesheet_delete_style (const string& stylesheet, const string& marker)
// Deletes a style.
{
  string filename;
  filename = directories_get_stylesheets ();
  filename = gw_build_filename (filename, stylesheet, marker + ".xml");
  unlink (filename.c_str());  
}


void stylesheet_new_style (const string& stylesheet, const string& marker, const string& name, const string& info)
// Adds a new style.
{
  string directory;
  directory = directories_get_stylesheets ();
  directory = gw_build_filename (directory, stylesheet);
  Style style (directory, marker, true);
  style.name = name;
  style.info = info;
}


void stylesheet_save_style (const string& stylesheet, const string& marker,
                            const string& name, const string& info,
                            StyleType type, double fontsize, int fontpercentage,
                            const string& italic, const string& bold, 
                            const string& underline, const string& smallcaps,
                            bool superscript, string justification,
                            double spacebefore, double spaceafter,
                            double leftmargin, double rightmargin,
                            double firstlineindent, bool spancolumns)
// Saves style information.
{
  string directory;
  directory = directories_get_stylesheets ();
  directory = gw_build_filename (directory, stylesheet);
  Style style (directory, marker, true);
  style.name = name;
  style.info = info;
  style.type = type;
  style.fontsize = fontsize;
  style.fontpercentage = fontpercentage;
  style.italic = italic;
  style.bold = bold;
  style.underline = underline;
  style.smallcaps = smallcaps;
  style.superscript = superscript;
  style.justification = justification;
  style.spacebefore = spacebefore;
  style.spaceafter = spaceafter;
  style.leftmargin = leftmargin;
  style.rightmargin = rightmargin;
  style.firstlineindent = firstlineindent;
  style.spancolumns = spancolumns;
}


int stylesheet_style_get_pointer (const vector<Style>& styles, const string& marker)
// Returns a pointer to "styles" which describes "marker".
// Or -1 if not found.
{
  // Search marker and return pointer.
  for (unsigned int i = 0; i < styles.size(); i++)
    if (marker == styles[i].marker)
      return i;
  // Ok, marker not found.
  return -1;
}


void stylesheet_initial_check ()
{
  // Get all available stylesheets.
  vector<string> stylesheets;
  stylesheet_get_ones_available (stylesheets);
  // No stylesheets available? Create standard one.
  if (stylesheets.size() == 0) {
    stylesheet_create_new (STANDARDSHEET);
    stylesheets.push_back (STANDARDSHEET);
  }
  // Update stylesheets to xml notation.
  for (unsigned int i = 0; i < stylesheets.size(); i++) {
    // Get the available old styles.
    string directory;
    directory = directories_get_stylesheets ();
    directory = gw_build_filename (directory, stylesheets[i]);
    ReadFiles rf (directory, "", "");
    for (unsigned int i = 0; i < rf.files.size(); i++) {
      if (rf.files[i].find ("xml") == string::npos) {
        // Storage for the data extracted from the old style file.
        string name;
        string info;
        double fontsize;
        bool italics;
        bool bold;
        bool superscript;
        string justification;
        double spacebefore;
        double spaceafter;
        double leftmargin;
        double rightmargin;
        double firstlineindent;
        bool spancolumns;
        StyleType type;
        // Read the style file. Add the bits that are missing.
        string filename = gw_build_filename (directory, rf.files[i]);
        ReadText rc (filename, true);
        for (unsigned int i = 0; i <= 100; i++) {
          switch (i) {
            case 2:
            {
              if (i < rc.lines.size ()) {
                name = rc.lines[i];
              }
              else {
                name = "Marker";
              }
              break;
            }
            case 3:
            {
              if (i < rc.lines.size ()) {
                info = rc.lines[i];
              }
              else
              {
                info = "Marker";
              }
              break;
            }
            case 5:
            {
              if (i < rc.lines.size ()) {
                fontsize = convert_to_double (rc.lines[i]);
              }
              else
              {
                fontsize = 12;
              }  
              break;
            }
            case 6:
            {
              if (i < rc.lines.size ()) {
                italics = convert_to_bool (rc.lines[i]);
              }
              else
              {
                italics = false;
              }
              break;
            }
            case 7:
            {  
              if (i < rc.lines.size ()) {
                bold = convert_to_bool (rc.lines[i]);
              }
              else
              {
                bold = false;
              }
              break;
            }
            case 8:
            {
              if (i < rc.lines.size ()) {
                superscript = convert_to_bool (rc.lines[i]);
              }
              else
              {
                superscript = false;
              }
              break;
            }
            case 10:
            {
              if (i < rc.lines.size ()) {
                justification = rc.lines[i];
              }
              else
              { 
                justification = "left";
              }
              break;
            }
            case 11:
            {
              if (i < rc.lines.size ()) {
                spacebefore = convert_to_double (rc.lines[i]);
              }
              else
              {
                spacebefore = 0;
              }
              break;
            }
            case 12:
            { 
              if (i < rc.lines.size ()) {
                spaceafter = convert_to_double (rc.lines[i]);
              }
              else
              {
                spaceafter = 0;
              }
              break;
            }
            case 13:
            {
              if (i < rc.lines.size ()) {
                leftmargin = convert_to_double (rc.lines[i]);
              }
              else
              {
                leftmargin = 0;
              }
              break;
            }
            case 14:
            {
              if (i < rc.lines.size ()) {
                rightmargin = convert_to_double (rc.lines[i]);
              }
              else
              {
                rightmargin = 0;
              }
              break;
            }
            case 15:
            {
              if (i < rc.lines.size ()) {
                firstlineindent = convert_to_double (rc.lines[i]);
              }
              else
              {
                firstlineindent = 0;
              }
              break;
            }
            case 16:
            {
              if (i < rc.lines.size ()) {
                spancolumns = convert_to_bool (rc.lines[i]);
              }
              else
              {
                spancolumns = false;
              }
              break;
            }
            case 17:
            {
              if (i < rc.lines.size ()) {
                type = StyleType (convert_to_int (rc.lines[i]));
              }
              else
              {
                type = stInlineText;
              }
              break;
            }
          }
        }
        // Convert the file to xml notation.
        vector<string> xml_lines;
        xml_lines.push_back ("<bibledit-style>");
        string s;
        s = "  <name>" + name + "</name>";
        xml_lines.push_back (s);
        s = "  <description>" + info + "</description>";
        xml_lines.push_back (s);
        s = "  <fontsize>" + convert_to_string (fontsize) + "</fontsize>";
        xml_lines.push_back (s);
        s = "  <italic>" + convert_to_string (italics) + "</italic>";
        xml_lines.push_back (s);
        s = "  <bold>" + convert_to_string (bold) + "</bold>";
        xml_lines.push_back (s);
        s = "  <superscript>" + convert_to_string (superscript) + "</superscript>";
        xml_lines.push_back (s);
        s = "  <justification>" + justification + "</justification>";
        xml_lines.push_back (s);
        s = "  <space-before>" + convert_to_string (spacebefore) + "</space-before>";
        xml_lines.push_back (s);
        s = "  <space-after>" + convert_to_string (spaceafter) + "</space-after>";
        xml_lines.push_back (s);
        s = "  <margin-left>" + convert_to_string (leftmargin) + "</margin-left>";
        xml_lines.push_back (s);
        s = "  <margin-right>" + convert_to_string (rightmargin) + "</margin-right>";
        xml_lines.push_back (s);
        s = "  <indent-firstline>" + convert_to_string (firstlineindent) + "</indent-firstline>";
        xml_lines.push_back (s);
        s = "  <columns-span>" + convert_to_string (spancolumns) + "</columns-span>";
        xml_lines.push_back (s);
        switch (type) {
          case stIdentifier:      { s = "identifier"; break; }
          case stComment:         { s = "comment"; break; }
          case stRunningHeader:   { s = "running-header"; break; }
          case stStartsParagraph: { s = "starts-paragraph"; break; }
          case stInlineText:      { s = "inline-text"; break; }
          case stChapterNumber:   { s = "chapter-number"; break; }
          case stVerseNumber:     { s = "verse-number"; break; }
        }
        s = "  <style-type>" + s + "</style-type>";
        xml_lines.push_back (s);
        xml_lines.push_back ("</bibledit-style>");
        write_lines (filename + ".xml", xml_lines);
        // Remove old style file.
        unlink (filename.c_str());
      }
    }
  }
}
