/*
** Copyright (C) 2003-2005 The Free Software Foundation.
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/


#include "libraries.h"
#include "utilities.h"
#include "usfm.h"
#include "usfmtools.h"


ustring usfm_extract (ustring & line)
{
  /*
     This returns the usfm from the line, e.g. \id.
     The usfm is removed from the line.
   */
  ustring s = trim (line);
  ustring returnvalue = "";
  if (s.length () > 1)
    {
      if (s.substr (0, 1) == "\\")
        {
          unsigned int endposition = s.find_first_of (" \\", 1);
          if (endposition == string::npos)
            {
              line = "";
              returnvalue = s;
            }
          else
            {
              returnvalue = s.substr (0, endposition);
              line.erase (0, endposition);
              line = trim (line);
            }
        }
    }
  return returnvalue;
}


ustring usfm_extract_marker (ustring & line)
{
  // Returns the usfm marker from the line, but without
  // the first backslash, so e.g. id.
  // The backslash and the usfm marker are removed
  // from the line.
  ustring returnvalue = usfm_extract (line);
  if (returnvalue.length () > 0)
    returnvalue.erase (0, 1);   // Remove backslash.
  return returnvalue;
}


int
usfm_search_marker_from (const ustring & line, unsigned int from)
{
  /* 
     Searches for any usfm (whether it exists or not).
     It starts searching at position "from".
     It returns the position where it found the usfm, or
     string::npos is not found.
   */
  return line.find ("\\", from);
}


unsigned int
usfm_search_marker_from_length (const ustring & line, unsigned int from)
{
  /*
     Returns the length of the usfm that is found at position
     "from".
   */
  ustring s;
  s = line.substr (from, line.length ());
  return usfm_extract (s).length ();
}


string
get_id_from_file (const string & filename)
{
  ReadText rt (filename, true);
  unsigned int max = 10;
  if (max >= rt.lines.size ())
    max = rt.lines.size ();
  for (unsigned int i = 0; i < max; i++) {
    ustring line = trim (rt.lines[i]);
    ustring marker = usfm_extract_marker (line);
    if (usfm_is_id (marker)) {
      try 
      {
        return line.substr (0, 3);
      }
      catch (exception & ex)
      {
        cerr << ex.what () << endl;
      }
    }
  }
  return "";
}


void split_line_into_usfm_and_text (const ustring& line, ustring& marker, ustring& text)
/*
Splits all text into marker and text. 
E.g. "\v 1 In the beginning ..." is split into:
- \v 1 
- In the beginning ...
*/
{
  // Do the first split.
  text = line;
  marker = usfm_extract_marker (text);
  // If this line did not have any usfm, we're through.
  if (marker.empty())
    return;
  // Certain usfms, like for chapter and verse, need to have the numbers too.
  if (usfm_is_chapter (marker) || usfm_is_verse (marker)) {
    string number;
    number = number_in_string (text);
    text.erase (0, number.length());
    text = trim (text);
    marker.append (" ");
    marker.append (number);
  }
  // The backslash was extracted, so put it back.
  marker.insert(0, "\\");
}


void usfm_handle_inline_text (ustring& line, UsfmInlineMarkers * inline_markers, XmlFoBlock * block)
// Deal with inline text.
{
  /*
  Strategy:
  1. Usfm text that has the opening marker, but not the closing marker, may result
     in malformed xslfo output. To ensure this does not occur, the opening tag
     is only placed in the xslfo output when the closing usfm is found too.
  2. For inline text markers, the space following the opening marker is
     part of the marker, and not part of the word.
  */
  // Go through all markers, and look for them in the line.
  for (unsigned int i = 0; i < inline_markers->opening_markers.size(); i++) {
    // Allow for a marker to occur more than once in the line.
    bool still_working = true;
    while (still_working) {
      string::size_type begin_position;
      begin_position = line.find(inline_markers->opening_markers[i]);
      if (begin_position != string::npos) {
        string::size_type end_position;
        end_position = line.find(inline_markers->closing_markers[i]);
        if (end_position != string::npos) {
          // At this stage we have both the opening and closing marker, so
          // we're safe to replace the markers with the xslfo tags.
          string::size_type position;
          string::size_type length;
          position = begin_position;
          length = inline_markers->opening_markers[i].length();
          string openingmarkers = inline_markers->opening_xslfo_markup (block, i);
          line.replace (position, length, openingmarkers);
          position = line.find(inline_markers->closing_markers[i]);
          length = inline_markers->closing_markers[i].length();
          line.replace (position, length, inline_markers->closing_xslfo_markup);
        } else still_working = false;
      } else still_working = false;
    }
  }
}


#define NUMBER_OF_MARKERS_IdentificationInformation 9
char *USFM_MARKERS_IdentificationInformation [NUMBER_OF_MARKERS_IdentificationInformation] =
{ 
  "id", "ide", "rem", "h", "h1", "h2", "h3", "toc1", "toc2"
};
#define NUMBER_OF_MARKERS_IntroductionElements 17
char *USFM_MARKERS_IntroductionElements [NUMBER_OF_MARKERS_IntroductionElements] =
{ 
  "imt", "imt1", "imt2", "imt3", "imt4", "is", "is1", "is2", "is3", "us4", "ip", "ipi", "im", "imi", "imq", "ipr", "ipq"
};
#define NUMBER_OF_MARKERS_IntroductionElements2 16
char *USFM_MARKERS_IntroductionElements2 [NUMBER_OF_MARKERS_IntroductionElements2] =
{ 
  "ib", "iex", "iq", "iq1", "iq2", "iq3", "iq4", "iot", "io", "io1", "io2", "io3", "io4", "ior", "imte", "ie"
};
#define NUMBER_OF_MARKERS_TitlesHeadingsAndLabels 15
char *USFM_MARKERS_TitlesHeadingsAndLabels [NUMBER_OF_MARKERS_TitlesHeadingsAndLabels] =
{ 
  "mt", "mt1", "mt2", "mt3", "mt4", "mte", "mte1", "mte2", "mte3", "mte4", "ms", "ms1", "ms2", "ms3", "ms4"
};
#define NUMBER_OF_MARKERS_TitlesHeadingsAndLabels2 10
char *USFM_MARKERS_TitlesHeadingsAndLabels2 [NUMBER_OF_MARKERS_TitlesHeadingsAndLabels2] =
{ 
  "mr", "s", "s1", "s2", "s3", "s4", "sr", "r", "d", "sp"
};
#define NUMBER_OF_MARKERS_ChaptersAndVerses 8
char *USFM_MARKERS_ChaptersAndVerses [NUMBER_OF_MARKERS_ChaptersAndVerses] =
{ 
  "c", "ca", "cl", "cp", "cd", "v", "va", "vp"
};
#define NUMBER_OF_MARKERS_ParagraphElements 27
char *USFM_MARKERS_ParagraphElements [NUMBER_OF_MARKERS_ParagraphElements] =
{ 
  "p", "m", "pmo", "pm", "pmc", "pmr", "pi", "pi1", "pi2", "pi3", "pi4", "mi", "nb", "cls", "li", "li1", "li2", "li3", "li4", "b", "pc", "pr", "ph", "ph1", "ph2", "ph3", "ph4"
};
#define NUMBER_OF_MARKERS_PoetryElements 16
char *USFM_MARKERS_PoetryElements [NUMBER_OF_MARKERS_PoetryElements] =
{ 
  "q", "q1", "q2", "q3", "q4", "qr", "qc", "qs", "qa", "qac", "qm", "qm1", "qm2", "qm3", "qm4", "b"
};
#define NUMBER_OF_MARKERS_TableElements 25
char *USFM_MARKERS_TableElements [NUMBER_OF_MARKERS_TableElements] =
{ 
  "tr", "th", "th1", "th2", "th3", "th4", "th5", "thr", "thr1", "thr2", "thr3", "thr4", "thr5", "tc", "tc1", "tc2", "tc3", "tc4", "tc5", "tcr", "tcr1", "tcr2", "tcr3", "tcr4", "tcr5"
};
#define NUMBER_OF_MARKERS_Footnotes 12
char *USFM_MARKERS_Footnotes [NUMBER_OF_MARKERS_Footnotes] =
{ 
  "f", "fe", "fr", "fk", "fq", "fqa", "fl", "fp", "fv", "ft", "fdc", "fm"
};
#define NUMBER_OF_MARKERS_CrossReferences 6
char *USFM_MARKERS_CrossReferences [NUMBER_OF_MARKERS_CrossReferences] =
{ 
  "x", "xo", "xk", "xq", "xt", "xdc"
};
#define NUMBER_OF_MARKERS_ExtendedStudyNotes 9
char *USFM_MARKERS_ExtendedStudyNotes [NUMBER_OF_MARKERS_ExtendedStudyNotes] =
{ 
  "env", "enw", "enk", "enc", "cat1", "cat2", "cat3", "cat4", "cat5"
};
#define NUMBER_OF_MARKERS_SpecialTextAndCharacterStyles 13
char *USFM_MARKERS_SpecialTextAndCharacterStyles [NUMBER_OF_MARKERS_SpecialTextAndCharacterStyles] =
{ 
  "qt", "nd", "tl", "dc", "bk", "sig", "pn", "wj", "k", "sis", "ord", "add", "lit"
};
#define NUMBER_OF_MARKERS_SpecialTextAndCharacterStyles2 14
char *USFM_MARKERS_SpecialTextAndCharacterStyles2 [NUMBER_OF_MARKERS_SpecialTextAndCharacterStyles2] =
{ 
  "no", "bd", "it", "bdit", "em", "sc", "pb", "fig", "pro", "w", "wh", "wg", "ndx", "sls"
};
#define NUMBER_OF_MARKERS_PeripheralMaterials 15
char *USFM_MARKERS_PeripheralMaterials [NUMBER_OF_MARKERS_PeripheralMaterials] =
{ 
  "pub", "toc", "pref", "intro", "conc", "glo", "idx", "maps", "cov", "spine", "k1", "k2", "p1", "p2", "pubinfo"
};
/*
Other markers that may occur in the text.
/restore
  This is a marker inserted by Paratext when restoring a Paratext project backup
  file, if the switch "Append this description to every retored file" is checked
  in the Restore dialogue. And so, it needs to be present in the stylesheet 
  (so that checking tools do not report a false error), but it is not really 
  part of USFM.
*/


void usfm_categorize_markers (vector<string>& markers, vector<UsfmCategories>& categories)
// Reorders markers and puts them in the right category.
{
  // Put all markers into a set, from where they will be ordered.
  // After ordering known markers, any remaining markers will be put in 
  // category Other.
  set<string> set_of_markers (markers.begin(), markers.end());
  // Clear markers and categories. Here we'll put the result of ordering them.
  markers.clear();
  categories.clear();
  for (unsigned int i = 0; i < NUMBER_OF_MARKERS_IdentificationInformation; i++) {
    set<string>::const_iterator found_position;
    found_position = set_of_markers.find (USFM_MARKERS_IdentificationInformation[i]);
    if (found_position != set_of_markers.end()) {
      markers.push_back (USFM_MARKERS_IdentificationInformation[i]);
      categories.push_back (ucIdentificationInformation);
      set_of_markers.erase (found_position);
    }
  }
  for (unsigned int i = 0; i < NUMBER_OF_MARKERS_IntroductionElements; i++) {
    set<string>::const_iterator found_position;
    found_position = set_of_markers.find (USFM_MARKERS_IntroductionElements[i]);
    if (found_position != set_of_markers.end()) {
      markers.push_back (USFM_MARKERS_IntroductionElements[i]);
      categories.push_back (ucIntroductionElements);
      set_of_markers.erase (found_position);
    }
  }
  for (unsigned int i = 0; i < NUMBER_OF_MARKERS_IntroductionElements2; i++) {
    set<string>::const_iterator found_position;
    found_position = set_of_markers.find (USFM_MARKERS_IntroductionElements2[i]);
    if (found_position != set_of_markers.end()) {
      markers.push_back (USFM_MARKERS_IntroductionElements2[i]);
      categories.push_back (ucIntroductionElements2);
      set_of_markers.erase (found_position);
    }
  }
  for (unsigned int i = 0; i < NUMBER_OF_MARKERS_TitlesHeadingsAndLabels; i++) {
    set<string>::const_iterator found_position;
    found_position = set_of_markers.find (USFM_MARKERS_TitlesHeadingsAndLabels[i]);
    if (found_position != set_of_markers.end()) {
      markers.push_back (USFM_MARKERS_TitlesHeadingsAndLabels[i]);
      categories.push_back (ucTitlesHeadingsAndLabels);
      set_of_markers.erase (found_position);
    }
  }
  for (unsigned int i = 0; i < NUMBER_OF_MARKERS_TitlesHeadingsAndLabels2; i++) {
    set<string>::const_iterator found_position;
    found_position = set_of_markers.find (USFM_MARKERS_TitlesHeadingsAndLabels2[i]);
    if (found_position != set_of_markers.end()) {
      markers.push_back (USFM_MARKERS_TitlesHeadingsAndLabels2[i]);
      categories.push_back (ucTitlesHeadingsAndLabels2);
      set_of_markers.erase (found_position);
    }
  }
  for (unsigned int i = 0; i < NUMBER_OF_MARKERS_ChaptersAndVerses; i++) {
    set<string>::const_iterator found_position;
    found_position = set_of_markers.find (USFM_MARKERS_ChaptersAndVerses[i]);
    if (found_position != set_of_markers.end()) {
      markers.push_back (USFM_MARKERS_ChaptersAndVerses[i]);
      categories.push_back (ucChaptersAndVerses);
      set_of_markers.erase (found_position);
    }
  }
  for (unsigned int i = 0; i < NUMBER_OF_MARKERS_ParagraphElements; i++) {
    set<string>::const_iterator found_position;
    found_position = set_of_markers.find (USFM_MARKERS_ParagraphElements[i]);
    if (found_position != set_of_markers.end()) {
      markers.push_back (USFM_MARKERS_ParagraphElements[i]);
      categories.push_back (ucParagraphElements);
      set_of_markers.erase (found_position);
    }
  }
  for (unsigned int i = 0; i < NUMBER_OF_MARKERS_PoetryElements; i++) {
    set<string>::const_iterator found_position;
    found_position = set_of_markers.find (USFM_MARKERS_PoetryElements[i]);
    if (found_position != set_of_markers.end()) {
      markers.push_back (USFM_MARKERS_PoetryElements[i]);
      categories.push_back (ucPoetryElements);
      set_of_markers.erase (found_position);
    }
  }
  for (unsigned int i = 0; i < NUMBER_OF_MARKERS_TableElements; i++) {
    set<string>::const_iterator found_position;
    found_position = set_of_markers.find (USFM_MARKERS_TableElements[i]);
    if (found_position != set_of_markers.end()) {
      markers.push_back (USFM_MARKERS_TableElements[i]);
      categories.push_back (ucTableElements);
      set_of_markers.erase (found_position);
    }
  }
  for (unsigned int i = 0; i < NUMBER_OF_MARKERS_Footnotes; i++) {
    set<string>::const_iterator found_position;
    found_position = set_of_markers.find (USFM_MARKERS_Footnotes[i]);
    if (found_position != set_of_markers.end()) {
      markers.push_back (USFM_MARKERS_Footnotes[i]);
      categories.push_back (ucFootnotes);
      set_of_markers.erase (found_position);
    }
  }
  for (unsigned int i = 0; i < NUMBER_OF_MARKERS_CrossReferences; i++) {
    set<string>::const_iterator found_position;
    found_position = set_of_markers.find (USFM_MARKERS_CrossReferences[i]);
    if (found_position != set_of_markers.end()) {
      markers.push_back (USFM_MARKERS_CrossReferences[i]);
      categories.push_back (ucCrossReferences);
      set_of_markers.erase (found_position);
    }
  }
  for (unsigned int i = 0; i < NUMBER_OF_MARKERS_ExtendedStudyNotes; i++) {
    set<string>::const_iterator found_position;
    found_position = set_of_markers.find (USFM_MARKERS_ExtendedStudyNotes[i]);
    if (found_position != set_of_markers.end()) {
      markers.push_back (USFM_MARKERS_ExtendedStudyNotes[i]);
      categories.push_back (ucExtendedStudyNotes);
      set_of_markers.erase (found_position);
    }
  }
  for (unsigned int i = 0; i < NUMBER_OF_MARKERS_SpecialTextAndCharacterStyles; i++) {
    set<string>::const_iterator found_position;
    found_position = set_of_markers.find (USFM_MARKERS_SpecialTextAndCharacterStyles[i]);
    if (found_position != set_of_markers.end()) {
      markers.push_back (USFM_MARKERS_SpecialTextAndCharacterStyles[i]);
      categories.push_back (ucSpecialTextAndCharacterStyles);
      set_of_markers.erase (found_position);
    }
  }
  for (unsigned int i = 0; i < NUMBER_OF_MARKERS_SpecialTextAndCharacterStyles2; i++) {
    set<string>::const_iterator found_position;
    found_position = set_of_markers.find (USFM_MARKERS_SpecialTextAndCharacterStyles2[i]);
    if (found_position != set_of_markers.end()) {
      markers.push_back (USFM_MARKERS_SpecialTextAndCharacterStyles2[i]);
      categories.push_back (ucSpecialTextAndCharacterStyles2);
      set_of_markers.erase (found_position);
    }
  }
  for (unsigned int i = 0; i < NUMBER_OF_MARKERS_PeripheralMaterials; i++) {
    set<string>::const_iterator found_position;
    found_position = set_of_markers.find (USFM_MARKERS_PeripheralMaterials[i]);
    if (found_position != set_of_markers.end()) {
      markers.push_back (USFM_MARKERS_PeripheralMaterials[i]);
      categories.push_back (ucPeripheralMaterials);
      set_of_markers.erase (found_position);
    }
  }
  // Having dealt with all known styles and put them in categories, dump
  // any remaining markers in the "Other" category.
  vector<string> other_markers (set_of_markers.begin(), set_of_markers.end());
  for (unsigned int i = 0; i < other_markers.size(); i++) {
    markers.push_back (other_markers[i]);
    categories.push_back (ucOtherElements);
  }
}


UsfmInlineMarkers::UsfmInlineMarkers (const Usfm& usfm)
// Creates a list of all inline markers, with their properties.
{
  // Go through all the styles and look for inline ones.
  for (unsigned int i = 0; i < usfm.styles.size(); i++) {
    if (usfm.styles[i].type == stInlineText) {
      // Inline marker found. Store opening marker.
      string s;
      s = "\\" + usfm.styles[i].marker + " ";
      opening_markers.push_back (s);
      // Store closing marker.      
      s = "\\" + usfm.styles[i].marker + "*";
      closing_markers.push_back (s);
      // Store the relevant properties of this style.
      fontpercentage.push_back(usfm.styles[i].fontpercentage);
      italic.push_back(usfm.styles[i].italic);
      bold.push_back(usfm.styles[i].bold);
      underline.push_back(usfm.styles[i].underline);
      smallcaps.push_back(usfm.styles[i].smallcaps);
      superscript.push_back(usfm.styles[i].superscript);
    }
  }
  // One and the same closing marker for everything.
  closing_xslfo_markup = "</fo:inline>";
}


UsfmInlineMarkers::~UsfmInlineMarkers ()
{
}



string UsfmInlineMarkers::opening_xslfo_markup (XmlFoBlock * block, int index)
// This one generates the opening xslfo markup.
// It uses 'index' for the index in the opening and closing markers, as there 
// can be many.
// If italic/bold/underline of the indexed marker is TOGGLE, it looks at the
// paragraph 'block' it is in to find out what markup should be given.
{
  string opening_xslfo_markup;
  // Deal with the fontpercentage.
  opening_xslfo_markup = "<fo:inline";
  opening_xslfo_markup.append (" font-size=\"");
  opening_xslfo_markup.append (convert_to_string (fontpercentage[index]));
  opening_xslfo_markup.append ("%\"");
  // Deal with italics.
  if (italic[index] == ON)
    opening_xslfo_markup.append (" font-style=\"italic\"");
  else if (italic[index] == OFF)
    opening_xslfo_markup.append (" font-style=\"normal\"");
  else if (italic[index] == INHERIT)
    opening_xslfo_markup.append (" font-style=\"inherit\"");
  else if (italic[index] == TOGGLE) {
    if (block) {
      if (block->italic)
        opening_xslfo_markup.append (" font-style=\"normal\"");
      else
        opening_xslfo_markup.append (" font-style=\"italic\"");
    }
  }
  // Deal with bold.
  if (bold[index] == ON)
    opening_xslfo_markup.append (" font-weight=\"bold\"");
  else if (bold[index] == OFF)
    opening_xslfo_markup.append (" font-weight=\"normal\"");
  else if (bold[index] == INHERIT)
    opening_xslfo_markup.append (" font-weight=\"inherit\"");
  else if (bold[index] == TOGGLE) {
    if (block) {
      if (block->bold)
        opening_xslfo_markup.append (" font-weight=\"normal\"");
      else
        opening_xslfo_markup.append (" font-weight=\"bold\"");
    }
  }
  // Deal with underline.
  if (underline[index] == ON)
    opening_xslfo_markup.append (" text-decoration=\"underline\"");
  else if (underline[index] == OFF)
    opening_xslfo_markup.append (" text-decoration=\"none\"");
  else if (underline[index] == INHERIT)
    opening_xslfo_markup.append (" text-decoration=\"inherit\"");
  else if (underline[index] == TOGGLE) {
    if (block) {
      if (block->underline)
        opening_xslfo_markup.append (" text-decoration=\"none\"");
      else
        opening_xslfo_markup.append (" text-decoration=\"underline\"");
    }
  }
  // Deal with smallcaps.
  if (smallcaps[index] == ON)
    opening_xslfo_markup.append (" font-variant=\"small-caps\"");
  else if (smallcaps[index] == OFF)
    opening_xslfo_markup.append (" font-variant=\"normal\"");
  else if (smallcaps[index] == INHERIT)
    opening_xslfo_markup.append (" font-variant=\"inherit\"");
  else if (smallcaps[index] == TOGGLE) {
    if (block) {
      if (block->smallcaps)
        opening_xslfo_markup.append (" font-variant=\"normal\"");
      else
        opening_xslfo_markup.append (" font-variant=\"small-caps\"");
    }
  }
  // Deal with superscript.
  if (superscript[index])
    opening_xslfo_markup.append (" vertical-align=\"super\"");
  // Close tag.
  opening_xslfo_markup.append (">");
  // Return markup.
  return opening_xslfo_markup;
}
