/*
** Copyright (C) 2003-2005 The Free Software Foundation.
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/


#include "libraries.h"
#include "utilities.h"
#include "worddatabase.h"
#include <config.h>
#include "sqlite3_plus.h"
#include "usfmtools.h"
#include "bible.h"
#include "log.h"
#include "book.h"
#include "usfm.h"
#include "scripture.h"
#include <glib.h>
#include "project.h"
#include "gwrappers.h"


#define TABLE_FILE_SIGNATURE "file_signature"
#define TABLE_LINES "lines"


Worddatabase::Worddatabase (int dummy)
{
}


Worddatabase::~Worddatabase ()
{
}


void Worddatabase::tick ()
{
  // Return if there are no objects to be dealt with.
  if (indexobjects.size() == 0)
    return;

  // If the object has been done, throw it out.
  if (indexobjects[0].done) {
    indexobjects.erase (indexobjects.begin());
    return;
  }
  
  // If the object is being done, return.
  if (indexobjects[0].beingdone)
    return;

  // Deal with this object.
  indexobjects[0].beingdone = true;
  switch (indexobjects[0].type) {
    case isCheckBook :
    {
      g_thread_create (GThreadFunc (check_book_thread), gpointer(this), false, NULL);
      break;
    }
    case isIndexChapter :
    {
      g_thread_create (GThreadFunc (index_chapter_thread), gpointer(this), false, NULL);
      break;
    }
    case isStoreSignature :
    {
      g_thread_create (GThreadFunc (store_signature_thread), gpointer (this), false, NULL);
      break;
    }
  }
}


void Worddatabase::initialize_books (const string& project)
{
  try
  {
    // Connect to the database and create the table(s) if needed.
    sqlite3::connection con (dbfilename (project).c_str ());
    int count;
    // Check for the table that stores the lines
    count = con.executeint32("select count(*) from sqlite_master where name='%s';", TABLE_LINES);
    if (count == 0)
      con.executenonquery("create table '%s' (book integer, chapter integer, verse integer, line text, casefolded text);", TABLE_LINES);
    // Check for the table that stores the file signature
    count = con.executeint32 ("select count(*) from sqlite_master where name='%s';",  TABLE_FILE_SIGNATURE);
    if (count == 0)
      con.executenonquery ("create table %s (filename text, signature text);", TABLE_FILE_SIGNATURE);
    con.close ();
    // Get books in project.
    Scripture scripture (project);
    // Check each book.
    for (unsigned int i = 0; i < scripture.paths.size (); i++) {
      IndexStorage indexstorage (project, scripture.paths[i]);
      indexobjects.push_back (indexstorage);
    }
  }
  catch (exception & ex)
  {
    cerr << "Error initializing: " << ex.what () << endl;
  }
}


void Worddatabase::update_chapter (const string& project, unsigned int book, unsigned int chapter, const vector <ustring>& lines)
{
  // Schedule this chapter for later execution. First come, first serve :)
  IndexStorage indexstorage (project, book, chapter, lines);
  indexobjects.push_back (indexstorage);
}


void Worddatabase::vacuum (const string& project)
// Vacuum the tables.
{
  try
  {
    sqlite3::connection con (dbfilename (project).c_str ());
    con.executenonquery("vacuum;");
    con.close ();
  }
  catch (exception & ex)
  {
    cerr << "Error vacuuming: " << ex.what () << endl;
  }
}


void Worddatabase::log (const string & message)
{
  log_message ("Indexer: " + message);
}


void Worddatabase::check_book_thread (gpointer data)
{
  ((Worddatabase *) data)->check_book ();
}


void Worddatabase::check_book ()
{
  try
  {
    // Connect to the database
    sqlite3::connection con (dbfilename (indexobjects[0].project).c_str ());
    // Get the actual signature for the file.
    string actual_signature;
    /* 
     * At first we got the signature through the ls -l command, but this caused 
     * a problem. It gives the modification time of the file with a resolution
     * of a minute. But if changes are applied to a file, and after the change
     * it still shows the same minute, then it did not get updates in the words
     * database. We now use the stat system call with higher accuracy of time.
     * This solves the problem 
     */
    struct stat statbuf;
    stat (indexobjects[0].filename.c_str(), &statbuf);
    actual_signature = convert_to_string (int(statbuf.st_size));
    actual_signature.append (convert_to_string (int(statbuf.st_mtime)));
    actual_signature.append (convert_to_string (int(statbuf.st_ctime)));
    // See if the filename exists in the table anyway.
    bool update = false;
    int count = con.executeint32 ("select count(*) from '%s' where filename='%s';", TABLE_FILE_SIGNATURE, indexobjects[0].filename.c_str ());
    if (count == 0) {
      // Doesn't exist, so schedule the file for updating the database.
      update = true;
    }
    if (!update) {
      // The filename is in the database - now check its signature.
      sqlite3::reader reader = con.executereader ("select signature from '%s' where filename='%s';", TABLE_FILE_SIGNATURE, indexobjects[0].filename.c_str ());
      if (reader.read ()) {
        string stored_signature = reader.getstring (0);
        if (stored_signature != actual_signature) {
          update = true;
        }
      }
      reader.close ();
    }
    if (update) {
      log ("Scheduled: " + indexobjects[0].filename);
      // Store data about this filename.
      Book wbook (indexobjects[0].filename, false, "");
      for (unsigned int chapter = 0; chapter <= wbook.number_of_chapters; chapter++) {
        vector <ustring> lines;
        wbook.get_chapter (chapter, lines);
        IndexStorage indexstorage (indexobjects[0].project, wbook.get_id_index (), chapter, lines);
        indexobjects.push_back (indexstorage);
      }
      // Schedule its signature for later storage.
      IndexStorage indexstorage (indexobjects[0].project, indexobjects[0].filename, actual_signature);
      indexobjects.push_back (indexstorage);
    }
    con.close ();
  }
  catch (exception & ex)
  {
    cerr << "Error checking: " << ex.what () << endl;
  }
  // Set object as done.
  indexobjects[0].done = true;
}


void Worddatabase::index_chapter_thread (gpointer data)
{
  ((Worddatabase *) data)->index_chapter ();
}


void Worddatabase::index_chapter ()
{
  log (index_to_biblebook (indexobjects[0].book) + " " + convert_to_string (indexobjects[0].chapter));
  try
  {
    /* 
      In order to avoid getting several times the same verse, when searching,
      we have to ensure only one line per verse is stored in the database.
      We first store all data in memory, and at the end add it to the database.
    */
    // Buffers where all data to be stored is collected.
    vector<unsigned int> book_to_store;
    vector<unsigned int> chapter_to_store;
    vector<string> verse_to_store;
    vector<ustring> line_to_store;
    vector<ustring> casefolded_line_to_store;
    string previous_verse;
    // Process each line.
    string currentverse = "0";
    for (unsigned int i = 0; i < indexobjects[0].lines.size (); i++) {
      ustring line (indexobjects[0].lines[i]);
      // Normalize the string for better result when comparing text.
      line = line.normalize ();
      // Extract the verse number.
      ustring string_line;
      string_line = line.raw ();
      ustring marker;
      marker = usfm_extract_marker (string_line);
      if (usfm_is_verse (marker))
        currentverse = number_in_string (string_line);
      // Any apostrophe (') should be prefixed by another one.
      line = double_apostrophy (line);
      ustring casefolded (line);
      casefolded = casefolded.casefold ();
      // Store the data in the buffers.
      if (currentverse != previous_verse) {
        // We've a new verse, so add it to the end of the buffers.
        previous_verse = currentverse;
        book_to_store.push_back(indexobjects[0].book);
        chapter_to_store.push_back(indexobjects[0].chapter);
        verse_to_store.push_back(currentverse);
        line_to_store.push_back(line);
        casefolded_line_to_store.push_back(casefolded);
      } else {
        // We've the same verse, so add it to the last items in the buffers.
        unsigned int last_item = line_to_store.size() - 1;
        line_to_store[last_item].append (" " + line);
        casefolded_line_to_store[last_item].append (" " + casefolded);
      }
    }
    // Connect and start transaction
    sqlite3::connection con (dbfilename (indexobjects[0].project).c_str ());
    con.executenonquery ("begin;");
    // Delete previous data (if any);
    con.executenonquery ("delete from '%s' where book = %d and chapter = %d;", TABLE_LINES, indexobjects[0].book, indexobjects[0].chapter);
    // Put data from the buffers in the database.
    for (unsigned int i = 0; i < line_to_store.size(); i++) {
      con.executenonquery ("insert into '%s' values (%d, %d, %s, '%s', '%s');", TABLE_LINES, book_to_store[i], chapter_to_store[i], verse_to_store[i].c_str(), line_to_store[i].c_str(), casefolded_line_to_store[i].c_str());
    }
    con.executenonquery ("commit;");
    con.close ();
  }
  catch (exception & ex)
  {
    cerr << "Error updating: " << ex.what () << endl;
  }
  // Set object as done.
  indexobjects[0].done = true;
}


void Worddatabase::delete_book (const string& project, unsigned int book)
{
  try
  {
    sqlite3::connection con (dbfilename (project).c_str ());
    con.executenonquery ("begin;");
    con.executenonquery ("delete from '%s' where book = %d;", TABLE_LINES, book);
    con.executenonquery ("commit;");
    con.close ();
  }
  catch (exception & ex)
  {
    cerr << "Error deleting: " << ex.what () << endl;
  }
}


string Worddatabase::dbfilename (const string& project)
{
  // Returns full path and name of the word database.
  string s = project + "_wordsdatabase.sqlite3";
  return gw_build_filename (PACKAGE_LOCALSTATE_DIR, s);
}


void Worddatabase::store_signature_thread (gpointer data)
{
  ((Worddatabase *) data)->store_signature ();
}


void Worddatabase::store_signature ()
{
  // Store signature of the book.
  try
  {
    sqlite3::connection con (dbfilename (indexobjects[0].project).c_str ());
    con.executenonquery ("begin;");
    con.executenonquery ("delete from '%s' where filename = '%s';", TABLE_FILE_SIGNATURE, indexobjects[0].filename.c_str ());
    con.executenonquery ("insert into '%s' values('%s', '%s');", TABLE_FILE_SIGNATURE, indexobjects[0].filename.c_str (), indexobjects[0].signature.c_str ());
    con.executenonquery ("commit;");
    con.close ();
  }
  catch (exception & ex)
  {
    cerr << "Error signing: " << ex.what () << endl;
  }
  indexobjects[0].done = true;
}


bool Worddatabase::is_ready ()
{
  return (indexobjects.size() == 0);
}
